import collections
import sys
import os
import json
from datetime import datetime, date

from snplib import load_snp_data, get_duration_end_date, get_series_stats, get_reinvestment_dates, get_investment_window, cluster_data

# the structure of data is:
#   date, price, earnings

spd_list = []
spd_dict = collections.OrderedDict()

def capital_gain(start_price, end_price):
    return (end_price - start_price)

def total_dividend(dividend_series):
    return sum([yd/12 for yd in dividend_series])

def total_return(investment_series):
    start_price, end_price = investment_series[0][1], investment_series[-1][1]
    capgain = capital_gain(start_price, end_price)
    divgain = total_dividend([x[2] for x in investment_series])
    return (capgain + divgain) / start_price

def compute_return_period(start, duration_months, with_dividends=True):
    end = get_duration_end_date(start, months=duration_months)
    if not with_dividends:
        start_price = spd_dict[start][0]
        end_price = spd_dict[end][0]
        return capital_gain(start_price, end_price)/start_price
    
    active_investment = get_investment_window(start, end, spd_list)
    if not active_investment or active_investment[-1][0] != end:
        return None
    return total_return(active_investment)

def compute_return_period_with_reinvestment(start, duration_months, reinvestment_period_months=None, with_dividends=True):
    # divide and conquer:
    # break down the investment into multiple sub-investments
    # return sum of returns of sub-investments
    if not reinvestment_period_months:
        return compute_return_period(start, duration_months, with_dividends)
    inv_end = get_duration_end_date(start, months=duration_months)
    sub_investment_returns = []
    for invdate in get_reinvestment_dates(start, duration_months, reinvestment_period_months):
        inv_duration = round((inv_end - invdate).days / 30.3)
        ret = compute_return_period(invdate, inv_duration, with_dividends=with_dividends)
        if ret is None:
            break
        sub_investment_returns.append(ret)
    if not sub_investment_returns:
        return None
    return sum(sub_investment_returns)/len(sub_investment_returns)

def find_periods_with_return(duration_months=12, min_return=0.0):
    loss_dates = []
    end_date = get_duration_end_date(max(spd_dict.keys()), months=-duration_months)
    for dt, value in spd_dict.items():
        if dt >= end_date:
            break
        ret = compute_return_period(dt, duration_months=duration_months)
        if ret < min_return:
            loss_dates.append([dt, ret])
    return loss_dates

def all_returns_raw(duration_months=12, reinvestment_period_months=None, with_dividends=True):
    dates = [d for d, _, _ in spd_list]
    retlist = []
    for d in dates:
        val = compute_return_period_with_reinvestment(d, duration_months, with_dividends=with_dividends, reinvestment_period_months=reinvestment_period_months)
        if val is None:
            break
        retlist.append(val)
    return retlist

def all_returns(duration_months=12, reinvestment_period_months=None, with_dividends=True):
    return get_series_stats(all_returns_raw(duration_months, reinvestment_period_months, with_dividends), precision=1)

def return_all_periods(reinvestment_period_months=None):
    duration_years = [1, 5, 10, 20, 35]
    return [(dy, all_returns(duration_months=12*dy, reinvestment_period_months=reinvestment_period_months)) for dy in duration_years]

def return_all_periods_with_reinvestment():
    reinvestment_period_months = [None, 12, 3, 1]
    res = dict()
    for reinv_pmo in reinvestment_period_months:
        res[reinv_pmo] = return_all_periods(reinvestment_period_months=reinv_pmo)
    return res

def scatterplot_all_returns_by_period(reinvestment_period_months=None, with_dividends=True):
    out = dict()
    duration_step_months = 3
    for durmonths in range(duration_step_months, 35*12, duration_step_months):
        data = all_returns_raw(durmonths, reinvestment_period_months, with_dividends)
        out[durmonths] = data
    return out

def output_scatterplot_data(rawdata):
    # output dataset 0: scatterplot
    stats = []
    stats_seq = ['mean', 'median_low', 'min', 'max', 'mode', 'stdev']
    dataset = []
    for period, series in rawdata.items():
        sts = get_series_stats(series, quantiles=50)
        qs = [sts['quantiles'][i] for i in [0, -1]]
        stats.append([period/12] + [sts[k] for k in stats_seq] + qs)
        for val, occurrences in cluster_data(series, transformf=lambda x: 100*x, bins=sts['quantiles']):
            dataset.append([period/12, val, occurrences])
    print(json.dumps(dataset))
    # output dataset 1: statistics over scatterplot
    print(json.dumps(stats))

def main():
    global spd_list
    global spd_dict
    try:
        filename = sys.argv[1]
    except:
        filename = os.path.join(os.path.dirname(__file__), 'snp500.csv')
    spd_list = load_snp_data(filename)
    # filter in only items from 1926: when S&P 500 was introduced.
    # I don't trust the data beforehand.
    spd_list = get_investment_window(date(1927, 1, 1), spd_list[-1][0], spd_list)
    spd_dict = {d: [p, e] for d, p, e in spd_list}

    cmd = sys.argv[2]

    if cmd == 'return':
        # some examples:
        # print the Total Return for a period of 35 years when entering the market on January 1, 1930 - and reinvesting every month
        # print(compute_return_period_with_reinvestment(date(1930, 1, 1), 35*12, 1))
        start_date = datetime([int(x) for x in sys.argv[3].split('-')]).date()
        duration_months = int(sys.argv[4])
        print(compute_return_period_with_reinvestment(start_date, duration_months))
    elif cmd == 'dates_lower_return':
        # print the Total Return for a period of 12 months when entering the market on January 1, 2017 - as a one-off investment
        # print(compute_return_period(date(2017, 1, 1), 12))
        expected_return = float(sys.argv[3])
        duration_months = int(sys.argv[4])
        print(find_periods_with_return(duration_months, expected_return))
    elif cmd == 'scatterplot':
        # print the Total Return for every possible 12-month investment horizon across history
        # print(all_returns(duration_months=12))
        # for dy, res in return_all_periods():
        #     print(f"{dy} => {res}")
        # for period, res in return_all_periods_with_reinvestment().items():
        #     print(f"{period} => {res}")
        splot = scatterplot_all_returns_by_period()
        output_scatterplot_data(splot)
    else:
        print("Command not recognized. See source for help.")

if __name__ == '__main__':
    main()