from datetime import date, datetime, timedelta
import statistics as st
import collections

def parse_row(r):
    parts = r.split(',')
    dt_parts = parts[0].split('-')
    try:
        dt = date(*[int(x) for x in dt_parts])
    except:
        return None
    try:
        price = float(parts[1])
        dividend = float(parts[2])
    except:
        return None
    return dt, price, dividend

def load_snp_data(filename):
    with open(filename) as f:
        return load_snp_data_string(f.read())

def load_snp_data_string(s):
    dataset = []
    for row in s.split('\n'):
        d = parse_row(row.strip())
        if d is not None:
            dataset.append(d)
    return dataset

def get_duration_end_date(start: datetime, years=0, months=0, days=0):
    e = start
    # add months
    ey = e.year + years + (months//12)
    em = e.month + (months%12)
    if em > 12:
        ey += 1
        em -= 12
    elif em < 0:
        ey -= 1
        em += 12
    e = e.replace(year=ey, month=em)
    return e + timedelta(days=days)

def get_reinvestment_dates(entry_date, duration_months, reinvestment_period_months):
    dates = [entry_date]
    newdate = entry_date
    for _ in range(1, duration_months // reinvestment_period_months):
        newmo = newdate.month+(reinvestment_period_months%12)
        newyr = newdate.year+reinvestment_period_months//12
        while newmo > 12:
            newyr += 1
            newmo -= 12
        newdate = newdate.replace(year=newyr, month=newmo)
        dates.append(newdate)
    return dates

def get_series_stats(series, precision=None, percent=True, quantiles=10):
    if not series:
        return None
    res = {
        f.__name__:f(series) for f in [
            st.mean,
            st.median_low,
            min,
            max,
            st.mode,
            st.stdev
        ]
    }
    qs = st.quantiles(series, n=quantiles)
    if percent:
        res = {x:100*y for x, y in res.items()}
        qs = [100*x for x in qs]
    if precision:
        res = {x:round(y, precision) for x, y in res.items()}
        qs = [round(x, precision) for x in qs]
    res['quantiles'] = qs
    return res

def count_months_delta(start, end):
    s, e = start, end
    years = e.year - s.year
    months = e.month - s.month
    if e.month < s.month:
        years -= 1
        months += 12
    return 12*years + months

def get_investment_window(start, end, data):
    # The code below is a O(1) equivalent of:
    #   return list(filter(lambda d: start <= d[0] <= end, data))
    first = data[0][0]
    last = data[-1][0]
    if start < first or end > last:
        return None
    pos_start = count_months_delta(first, start)
    pos_end = count_months_delta(start, end) + pos_start
    try:
        assert data[pos_start][0] == start
        assert data[pos_end][0] == end
    except:
        raise ValueError(f"Cannot find boundaries {start}:{end}, got {pos_start}:{pos_end} => {data[pos_start][0]}:{data[pos_end][0]} ({count_months_delta(first, start)}/{count_months_delta(start, end)})")
    return data[pos_start:pos_end+1]

def smart_round(val, significant_digits=3):
    if val == int(val):
        ndec, nint = 0, len(str(int(val)))
    else:
        nint, ndec = [len(x) for x in str(val).split('.')]
    digits_to_remove = (nint + ndec) - significant_digits
    if digits_to_remove <= 0:
        return val
    if ndec > 0:
        val = round(val, ndec-digits_to_remove)
        digits_to_remove -= ndec
        ndec = 0
    if val == int(val):
        return int(val)
    return val

def bin_data(item, bin_thresholds, dmin, dmax):
    """Return the bin number and centroid where item belongs."""
    ncentroids = len(bin_thresholds)
    def bin_centroid(n):
        if n == 0:
            return (dmin+bin_thresholds[0])/2
        if n == ncentroids:
            return (dmax+bin_thresholds[-1])/2
        return (bin_thresholds[n]+bin_thresholds[n-1])/2
    for n, bt in enumerate(bin_thresholds):
        if item < bt:
            return n, bin_centroid(n)
    return ncentroids, bin_centroid(ncentroids)

def cluster_data(series, transformf=None, bins=None):
    if transformf is None:
        transformf = lambda x: x
    data = [transformf(val) for val in series]
    # make buckets
    dmin, dmax = min(data), max(data)
    if bins:
        data = [bin_data(val, bins, dmin, dmax)[1] for val in data]
    else:
        dstep = (dmax-dmin) / 30
        data = [dmin+dstep*((val-dmin)//dstep)+dstep/2 for val in data]
        data = [smart_round(val) for val in data]
    cnt = collections.Counter(data)
    tot = sum(cnt.values())
    return [(val, round((100*occurrences)/tot)) for val, occurrences in cnt.most_common()]


if __name__ == '__main__':
    # some tests
    assert smart_round(1.1111) == 1.11
    assert smart_round(1.1311) == 1.13
    assert smart_round(1000212123.1111) == 1000000000
    assert smart_round(1234, significant_digits=3) == 1230