from statistics import stdev
import unittest
from datetime import datetime, date

from snplib import load_snp_data_string, get_duration_end_date, get_reinvestment_dates, get_investment_window, count_months_delta, smart_round

class TestLoadData(unittest.TestCase):
    def test_load_snp_data_string_empty(self):
        self.assertEqual([], load_snp_data_string('  '))

    def test_load_snp_data_string_one(self):
        want = [(date(2021, 12, 18), 123.3, 0.23), ]
        self.assertEqual(want, load_snp_data_string('2021-12-18,123.3,0.23'))

    def test_load_snp_data_string_invalid_rows(self):
        want = [(date(2021, 12, 18), 123.3, 0.23), (date(2021, 12, 19), 33.1, 0.11), ]
        s = """2021-12-18,123.3,0.23
2021-12-18,123.3,a
2021-12-19,33.1,0.11"""
        self.assertEqual(want, load_snp_data_string(s))


class TestGetDurationEndDate(unittest.TestCase):
    def test_get_duration_end_date_type(self):
        start = datetime.now()
        self.assertEqual(get_duration_end_date(start), start)

    def test_get_duration_end_date_years(self):
        start = datetime.now().replace(year=1980, month=12, day=22)
        self.assertEqual(get_duration_end_date(start, years=2), start.replace(year=start.year+2))

    def test_get_duration_end_date_months(self):
        start = datetime.now().replace(year=1980, month=12, day=22)
        want_date = date(1983, 2, 22)
        self.assertEqual(get_duration_end_date(start, months=26).date(), want_date)

    def test_get_duration_end_date_combined(self):
        start = datetime.now().replace(year=1980, month=12, day=22)
        want_date = date(1992, 2, 26)
        self.assertEqual(get_duration_end_date(start, years=9, months=26, days=4).date(), want_date)


class TestGetReinvestmentDates(unittest.TestCase):
    def test_get_reinvestment_dates_none(self):
        start = date(1980, 4, 3)
        self.assertEqual([start], get_reinvestment_dates(start, duration_months=1, reinvestment_period_months=1))
        self.assertEqual([start], get_reinvestment_dates(start, duration_months=1, reinvestment_period_months=13))

    def test_get_reinvestment_dates_multiple(self):
        start = date(1980, 11, 3)
        have = get_reinvestment_dates(start, duration_months=3, reinvestment_period_months=1)
        self.assertEqual(len(have), 3)
        want = [start, date(1980, 12, 3), date(1981, 1, 3)]
        self.assertEqual(have, want)
    
    def test_get_reinvestment_dates_year(self):
        start = date(1980, 11, 3)
        have = get_reinvestment_dates(start, duration_months=12, reinvestment_period_months=12)
        want = [start]
        self.assertEqual(have, want)
    


class TestGetInvestmentWindow(unittest.TestCase):
    def setUp(self) -> None:
        d = [date(1871, 1, 1), date(1871, 2, 1), date(1871, 3, 1), date(1871, 4, 1), date(1871, 5, 1), date(1871, 6, 1), date(1871, 7, 1), date(1871, 8, 1), date(1871, 9, 1), date(1871, 10, 1), date(1871, 11, 1), date(1871, 12, 1), date(1872, 1, 1), date(1872, 2, 1), date(1872, 3, 1), date(1872, 4, 1), date(1872, 5, 1), date(1872, 6, 1), date(1872, 7, 1), date(1872, 8, 1), date(1872, 9, 1), date(1872, 10, 1), date(1872, 11, 1), date(1872, 12, 1), date(1873, 1, 1), date(1873, 2, 1), date(1873, 3, 1), date(1873, 4, 1), date(1873, 5, 1), date(1873, 6, 1), date(1873, 7, 1), date(1873, 8, 1), date(1873, 9, 1), date(1873, 10, 1), date(1873, 11, 1), date(1873, 12, 1), date(1874, 1, 1), date(1874, 2, 1), date(1874, 3, 1), date(1874, 4, 1), date(1874, 5, 1), date(1874, 6, 1), date(1874, 7, 1), date(1874, 8, 1), date(1874, 9, 1), date(1874, 10, 1), date(1874, 11, 1), date(1874, 12, 1), date(1875, 1, 1), date(1875, 2, 1), date(1875, 3, 1), date(1875, 4, 1), date(1875, 5, 1), date(1875, 6, 1), date(1875, 7, 1), date(1875, 8, 1), date(1875, 9, 1), date(1875, 10, 1), date(1875, 11, 1), date(1875, 12, 1), date(1876, 1, 1), date(1876, 2, 1), date(1876, 3, 1), date(1876, 4, 1), date(1876, 5, 1), date(1876, 6, 1), date(1876, 7, 1), date(1876, 8, 1), date(1876, 9, 1), date(1876, 10, 1), date(1876, 11, 1), date(1876, 12, 1), date(1877, 1, 1), date(1877, 2, 1), date(1877, 3, 1), date(1877, 4, 1), date(1877, 5, 1), date(1877, 6, 1), date(1877, 7, 1), date(1877, 8, 1), date(1877, 9, 1), date(1877, 10, 1), date(1877, 11, 1), date(1877, 12, 1), date(1878, 1, 1), date(1878, 2, 1), date(1878, 3, 1), date(1878, 4, 1), date(1878, 5, 1), date(1878, 6, 1), date(1878, 7, 1), date(1878, 8, 1), date(1878, 9, 1), date(1878, 10, 1), date(1878, 11, 1), date(1878, 12, 1), date(1879, 1, 1), date(1879, 2, 1), date(1879, 3, 1), date(1879, 4, 1), date(1879, 5, 1), date(1879, 6, 1), date(1879, 7, 1), date(1879, 8, 1), date(1879, 9, 1), date(1879, 10, 1), date(1879, 11, 1), date(1879, 12, 1), date(1880, 1, 1), date(1880, 2, 1), date(1880, 3, 1), date(1880, 4, 1), date(1880, 5, 1), date(1880, 6, 1), date(1880, 7, 1), date(1880, 8, 1), date(1880, 9, 1), date(1880, 10, 1), date(1880, 11, 1), date(1880, 12, 1), date(1881, 1, 1), date(1881, 2, 1), date(1881, 3, 1), date(1881, 4, 1), date(1881, 5, 1), date(1881, 6, 1), date(1881, 7, 1), date(1881, 8, 1), date(1881, 9, 1), date(1881, 10, 1), date(1881, 11, 1), date(1881, 12, 1), date(1882, 1, 1), date(1882, 2, 1), date(1882, 3, 1), date(1882, 4, 1), date(1882, 5, 1), date(1882, 6, 1), date(1882, 7, 1), date(1882, 8, 1), date(1882, 9, 1), date(1882, 10, 1), date(1882, 11, 1), date(1882, 12, 1), date(1883, 1, 1), date(1883, 2, 1), date(1883, 3, 1), date(1883, 4, 1), date(1883, 5, 1), date(1883, 6, 1), date(1883, 7, 1), date(1883, 8, 1), date(1883, 9, 1), date(1883, 10, 1), date(1883, 11, 1), date(1883, 12, 1), date(1884, 1, 1), date(1884, 2, 1), date(1884, 3, 1), date(1884, 4, 1), date(1884, 5, 1), date(1884, 6, 1), date(1884, 7, 1), date(1884, 8, 1), date(1884, 9, 1), date(1884, 10, 1), date(1884, 11, 1), date(1884, 12, 1), date(1885, 1, 1), date(1885, 2, 1), date(1885, 3, 1), date(1885, 4, 1), date(1885, 5, 1), date(1885, 6, 1), date(1885, 7, 1), date(1885, 8, 1), date(1885, 9, 1), date(1885, 10, 1), date(1885, 11, 1), date(1885, 12, 1), date(1886, 1, 1), date(1886, 2, 1), date(1886, 3, 1), date(1886, 4, 1), date(1886, 5, 1), date(1886, 6, 1), date(1886, 7, 1), date(1886, 8, 1), date(1886, 9, 1), date(1886, 10, 1), date(1886, 11, 1), date(1886, 12, 1), date(1887, 1, 1), date(1887, 2, 1), date(1887, 3, 1), date(1887, 4, 1), date(1887, 5, 1), date(1887, 6, 1), date(1887, 7, 1), date(1887, 8, 1), date(1887, 9, 1), date(1887, 10, 1), date(1887, 11, 1), date(1887, 12, 1), date(1888, 1, 1), date(1888, 2, 1), date(1888, 3, 1), date(1888, 4, 1), date(1888, 5, 1), date(1888, 6, 1), date(1888, 7, 1), date(1888, 8, 1), date(1888, 9, 1), date(1888, 10, 1), date(1888, 11, 1), date(1888, 12, 1), date(1889, 1, 1), date(1889, 2, 1), date(1889, 3, 1), date(1889, 4, 1), date(1889, 5, 1), date(1889, 6, 1), date(1889, 7, 1), date(1889, 8, 1), date(1889, 9, 1), date(1889, 10, 1), date(1889, 11, 1), date(1889, 12, 1), date(1890, 1, 1), date(1890, 2, 1), date(1890, 3, 1), date(1890, 4, 1), date(1890, 5, 1), date(1890, 6, 1), date(1890, 7, 1), date(1890, 8, 1), date(1890, 9, 1), date(1890, 10, 1), date(1890, 11, 1), date(1890, 12, 1), date(1891, 1, 1), date(1891, 2, 1), date(1891, 3, 1), date(1891, 4, 1), date(1891, 5, 1), date(1891, 6, 1), date(1891, 7, 1), date(1891, 8, 1), date(1891, 9, 1), date(1891, 10, 1), date(1891, 11, 1), date(1891, 12, 1), date(1892, 1, 1), date(1892, 2, 1), date(1892, 3, 1), date(1892, 4, 1), date(1892, 5, 1), date(1892, 6, 1), date(1892, 7, 1), date(1892, 8, 1), date(1892, 9, 1), date(1892, 10, 1), date(1892, 11, 1), date(1892, 12, 1), date(1893, 1, 1), date(1893, 2, 1), date(1893, 3, 1), date(1893, 4, 1), date(1893, 5, 1), date(1893, 6, 1), date(1893, 7, 1), date(1893, 8, 1), date(1893, 9, 1), date(1893, 10, 1), date(1893, 11, 1), date(1893, 12, 1), date(1894, 1, 1), date(1894, 2, 1), date(1894, 3, 1), date(1894, 4, 1), date(1894, 5, 1), date(1894, 6, 1), date(1894, 7, 1), date(1894, 8, 1), date(1894, 9, 1), date(1894, 10, 1), date(1894, 11, 1), date(1894, 12, 1), date(1895, 1, 1), date(1895, 2, 1), date(1895, 3, 1), date(1895, 4, 1), date(1895, 5, 1), date(1895, 6, 1), date(1895, 7, 1), date(1895, 8, 1), date(1895, 9, 1), date(1895, 10, 1), date(1895, 11, 1), date(1895, 12, 1), date(1896, 1, 1), date(1896, 2, 1), date(1896, 3, 1), date(1896, 4, 1), date(1896, 5, 1), date(1896, 6, 1), date(1896, 7, 1), date(1896, 8, 1), date(1896, 9, 1), date(1896, 10, 1), date(1896, 11, 1), date(1896, 12, 1), date(1897, 1, 1), date(1897, 2, 1), date(1897, 3, 1), date(1897, 4, 1), date(1897, 5, 1), date(1897, 6, 1), date(1897, 7, 1), date(1897, 8, 1), date(1897, 9, 1), date(1897, 10, 1), date(1897, 11, 1), date(1897, 12, 1), date(1898, 1, 1), date(1898, 2, 1), date(1898, 3, 1), date(1898, 4, 1), date(1898, 5, 1), date(1898, 6, 1), date(1898, 7, 1), date(1898, 8, 1), date(1898, 9, 1), date(1898, 10, 1), date(1898, 11, 1), date(1898, 12, 1), date(1899, 1, 1), date(1899, 2, 1), date(1899, 3, 1), date(1899, 4, 1), date(1899, 5, 1), date(1899, 6, 1), date(1899, 7, 1), date(1899, 8, 1), date(1899, 9, 1), date(1899, 10, 1), date(1899, 11, 1), date(1899, 12, 1), date(1900, 1, 1), date(1900, 2, 1), date(1900, 3, 1), date(1900, 4, 1), date(1900, 5, 1), date(1900, 6, 1), date(1900, 7, 1), date(1900, 8, 1), date(1900, 9, 1), date(1900, 10, 1), date(1900, 11, 1), date(1900, 12, 1), date(1901, 1, 1), date(1901, 2, 1), date(1901, 3, 1), date(1901, 4, 1), date(1901, 5, 1), date(1901, 6, 1), date(1901, 7, 1), date(1901, 8, 1), date(1901, 9, 1), date(1901, 10, 1), date(1901, 11, 1), date(1901, 12, 1), date(1902, 1, 1), date(1902, 2, 1), date(1902, 3, 1), date(1902, 4, 1), date(1902, 5, 1), date(1902, 6, 1), date(1902, 7, 1), date(1902, 8, 1), date(1902, 9, 1), date(1902, 10, 1), date(1902, 11, 1), date(1902, 12, 1), date(1903, 1, 1), date(1903, 2, 1), date(1903, 3, 1), date(1903, 4, 1), date(1903, 5, 1), date(1903, 6, 1), date(1903, 7, 1), date(1903, 8, 1), date(1903, 9, 1), date(1903, 10, 1), date(1903, 11, 1), date(1903, 12, 1), date(1904, 1, 1), date(1904, 2, 1), date(1904, 3, 1), date(1904, 4, 1), date(1904, 5, 1), date(1904, 6, 1), date(1904, 7, 1), date(1904, 8, 1), date(1904, 9, 1), date(1904, 10, 1), date(1904, 11, 1), date(1904, 12, 1), date(1905, 1, 1), date(1905, 2, 1), date(1905, 3, 1), date(1905, 4, 1), date(1905, 5, 1), date(1905, 6, 1), date(1905, 7, 1), date(1905, 8, 1), date(1905, 9, 1), date(1905, 10, 1), date(1905, 11, 1), date(1905, 12, 1), date(1906, 1, 1), date(1906, 2, 1), date(1906, 3, 1), date(1906, 4, 1), date(1906, 5, 1), date(1906, 6, 1), date(1906, 7, 1), date(1906, 8, 1), date(1906, 9, 1), date(1906, 10, 1), date(1906, 11, 1), date(1906, 12, 1), date(1907, 1, 1), date(1907, 2, 1), date(1907, 3, 1), date(1907, 4, 1), date(1907, 5, 1), date(1907, 6, 1), date(1907, 7, 1), date(1907, 8, 1), date(1907, 9, 1), date(1907, 10, 1), date(1907, 11, 1), date(1907, 12, 1), date(1908, 1, 1), date(1908, 2, 1), date(1908, 3, 1), date(1908, 4, 1), date(1908, 5, 1), date(1908, 6, 1), date(1908, 7, 1), date(1908, 8, 1), date(1908, 9, 1), date(1908, 10, 1), date(1908, 11, 1), date(1908, 12, 1), date(1909, 1, 1), date(1909, 2, 1), date(1909, 3, 1), date(1909, 4, 1), date(1909, 5, 1), date(1909, 6, 1), date(1909, 7, 1), date(1909, 8, 1), date(1909, 9, 1), date(1909, 10, 1), date(1909, 11, 1), date(1909, 12, 1), date(1910, 1, 1), date(1910, 2, 1), date(1910, 3, 1), date(1910, 4, 1), date(1910, 5, 1), date(1910, 6, 1), date(1910, 7, 1), date(1910, 8, 1), date(1910, 9, 1), date(1910, 10, 1), date(1910, 11, 1), date(1910, 12, 1), date(1911, 1, 1), date(1911, 2, 1), date(1911, 3, 1), date(1911, 4, 1), date(1911, 5, 1), date(1911, 6, 1), date(1911, 7, 1), date(1911, 8, 1), date(1911, 9, 1), date(1911, 10, 1), date(1911, 11, 1), date(1911, 12, 1), date(1912, 1, 1), date(1912, 2, 1), date(1912, 3, 1), date(1912, 4, 1), date(1912, 5, 1), date(1912, 6, 1), date(1912, 7, 1), date(1912, 8, 1), date(1912, 9, 1), date(1912, 10, 1), date(1912, 11, 1), date(1912, 12, 1), date(1913, 1, 1), date(1913, 2, 1), date(1913, 3, 1), date(1913, 4, 1), date(1913, 5, 1), date(1913, 6, 1), date(1913, 7, 1), date(1913, 8, 1), date(1913, 9, 1), date(1913, 10, 1), date(1913, 11, 1), date(1913, 12, 1), date(1914, 1, 1), date(1914, 2, 1), date(1914, 3, 1), date(1914, 4, 1), date(1914, 5, 1), date(1914, 6, 1), date(1914, 7, 1), date(1914, 8, 1), date(1914, 9, 1), date(1914, 10, 1), date(1914, 11, 1), date(1914, 12, 1), date(1915, 1, 1), date(1915, 2, 1), date(1915, 3, 1), date(1915, 4, 1), date(1915, 5, 1), date(1915, 6, 1), date(1915, 7, 1), date(1915, 8, 1), date(1915, 9, 1), date(1915, 10, 1), date(1915, 11, 1), date(1915, 12, 1), date(1916, 1, 1), date(1916, 2, 1), date(1916, 3, 1), date(1916, 4, 1), date(1916, 5, 1), date(1916, 6, 1), date(1916, 7, 1), date(1916, 8, 1), date(1916, 9, 1), date(1916, 10, 1), date(1916, 11, 1), date(1916, 12, 1), date(1917, 1, 1), date(1917, 2, 1), date(1917, 3, 1), date(1917, 4, 1), date(1917, 5, 1), date(1917, 6, 1), date(1917, 7, 1), date(1917, 8, 1), date(1917, 9, 1), date(1917, 10, 1), date(1917, 11, 1), date(1917, 12, 1), date(1918, 1, 1), date(1918, 2, 1), date(1918, 3, 1), date(1918, 4, 1), date(1918, 5, 1), date(1918, 6, 1), date(1918, 7, 1), date(1918, 8, 1), date(1918, 9, 1), date(1918, 10, 1), date(1918, 11, 1), date(1918, 12, 1), date(1919, 1, 1), date(1919, 2, 1), date(1919, 3, 1), date(1919, 4, 1), date(1919, 5, 1), date(1919, 6, 1), date(1919, 7, 1), date(1919, 8, 1), date(1919, 9, 1), date(1919, 10, 1), date(1919, 11, 1), date(1919, 12, 1), date(1920, 1, 1), date(1920, 2, 1), date(1920, 3, 1), date(1920, 4, 1), date(1920, 5, 1), date(1920, 6, 1), date(1920, 7, 1), date(1920, 8, 1), date(1920, 9, 1), date(1920, 10, 1), date(1920, 11, 1), date(1920, 12, 1), date(1921, 1, 1), date(1921, 2, 1), date(1921, 3, 1), date(1921, 4, 1), date(1921, 5, 1), date(1921, 6, 1), date(1921, 7, 1), date(1921, 8, 1), date(1921, 9, 1), date(1921, 10, 1), date(1921, 11, 1), date(1921, 12, 1), date(1922, 1, 1), date(1922, 2, 1), date(1922, 3, 1), date(1922, 4, 1), date(1922, 5, 1), date(1922, 6, 1), date(1922, 7, 1), date(1922, 8, 1), date(1922, 9, 1), date(1922, 10, 1), date(1922, 11, 1), date(1922, 12, 1), date(1923, 1, 1), date(1923, 2, 1), date(1923, 3, 1), date(1923, 4, 1), date(1923, 5, 1), date(1923, 6, 1), date(1923, 7, 1), date(1923, 8, 1), date(1923, 9, 1), date(1923, 10, 1), date(1923, 11, 1), date(1923, 12, 1), date(1924, 1, 1), date(1924, 2, 1), date(1924, 3, 1), date(1924, 4, 1), date(1924, 5, 1), date(1924, 6, 1), date(1924, 7, 1), date(1924, 8, 1), date(1924, 9, 1), date(1924, 10, 1), date(1924, 11, 1), date(1924, 12, 1), date(1925, 1, 1), date(1925, 2, 1), date(1925, 3, 1), date(1925, 4, 1), date(1925, 5, 1), date(1925, 6, 1), date(1925, 7, 1), date(1925, 8, 1), date(1925, 9, 1), date(1925, 10, 1), date(1925, 11, 1), date(1925, 12, 1), date(1926, 1, 1), date(1926, 2, 1), date(1926, 3, 1), date(1926, 4, 1), date(1926, 5, 1), date(1926, 6, 1), date(1926, 7, 1), date(1926, 8, 1), date(1926, 9, 1), date(1926, 10, 1), date(1926, 11, 1), date(1926, 12, 1), date(1927, 1, 1), date(1927, 2, 1), date(1927, 3, 1), date(1927, 4, 1), date(1927, 5, 1), date(1927, 6, 1), date(1927, 7, 1), date(1927, 8, 1), date(1927, 9, 1), date(1927, 10, 1), date(1927, 11, 1), date(1927, 12, 1), date(1928, 1, 1), date(1928, 2, 1), date(1928, 3, 1), date(1928, 4, 1), date(1928, 5, 1), date(1928, 6, 1), date(1928, 7, 1), date(1928, 8, 1), date(1928, 9, 1), date(1928, 10, 1), date(1928, 11, 1), date(1928, 12, 1), date(1929, 1, 1), date(1929, 2, 1), date(1929, 3, 1), date(1929, 4, 1), date(1929, 5, 1), date(1929, 6, 1), date(1929, 7, 1), date(1929, 8, 1), date(1929, 9, 1), date(1929, 10, 1), date(1929, 11, 1), date(1929, 12, 1), date(1930, 1, 1), date(1930, 2, 1), date(1930, 3, 1), date(1930, 4, 1), date(1930, 5, 1), date(1930, 6, 1), date(1930, 7, 1), date(1930, 8, 1), date(1930, 9, 1), date(1930, 10, 1), date(1930, 11, 1), date(1930, 12, 1), date(1931, 1, 1), date(1931, 2, 1), date(1931, 3, 1), date(1931, 4, 1), date(1931, 5, 1), date(1931, 6, 1), date(1931, 7, 1), date(1931, 8, 1), date(1931, 9, 1), date(1931, 10, 1), date(1931, 11, 1), date(1931, 12, 1), date(1932, 1, 1), date(1932, 2, 1), date(1932, 3, 1), date(1932, 4, 1), date(1932, 5, 1), date(1932, 6, 1), date(1932, 7, 1), date(1932, 8, 1), date(1932, 9, 1), date(1932, 10, 1), date(1932, 11, 1), date(1932, 12, 1), date(1933, 1, 1), date(1933, 2, 1), date(1933, 3, 1), date(1933, 4, 1), date(1933, 5, 1), date(1933, 6, 1), date(1933, 7, 1), date(1933, 8, 1), date(1933, 9, 1), date(1933, 10, 1), date(1933, 11, 1), date(1933, 12, 1), date(1934, 1, 1), date(1934, 2, 1), date(1934, 3, 1), date(1934, 4, 1), date(1934, 5, 1), date(1934, 6, 1), date(1934, 7, 1), date(1934, 8, 1), date(1934, 9, 1), date(1934, 10, 1), date(1934, 11, 1), date(1934, 12, 1), date(1935, 1, 1), date(1935, 2, 1), date(1935, 3, 1), date(1935, 4, 1), date(1935, 5, 1), date(1935, 6, 1), date(1935, 7, 1), date(1935, 8, 1), date(1935, 9, 1), date(1935, 10, 1), date(1935, 11, 1), date(1935, 12, 1), date(1936, 1, 1), date(1936, 2, 1), date(1936, 3, 1), date(1936, 4, 1), date(1936, 5, 1), date(1936, 6, 1), date(1936, 7, 1), date(1936, 8, 1), date(1936, 9, 1), date(1936, 10, 1), date(1936, 11, 1), date(1936, 12, 1), date(1937, 1, 1), date(1937, 2, 1), date(1937, 3, 1), date(1937, 4, 1), date(1937, 5, 1), date(1937, 6, 1), date(1937, 7, 1), date(1937, 8, 1), date(1937, 9, 1), date(1937, 10, 1), date(1937, 11, 1), date(1937, 12, 1), date(1938, 1, 1), date(1938, 2, 1), date(1938, 3, 1), date(1938, 4, 1), date(1938, 5, 1), date(1938, 6, 1), date(1938, 7, 1), date(1938, 8, 1), date(1938, 9, 1), date(1938, 10, 1), date(1938, 11, 1), date(1938, 12, 1), date(1939, 1, 1), date(1939, 2, 1), date(1939, 3, 1), date(1939, 4, 1), date(1939, 5, 1), date(1939, 6, 1), date(1939, 7, 1), date(1939, 8, 1), date(1939, 9, 1), date(1939, 10, 1), date(1939, 11, 1), date(1939, 12, 1), date(1940, 1, 1), date(1940, 2, 1), date(1940, 3, 1), date(1940, 4, 1), date(1940, 5, 1), date(1940, 6, 1), date(1940, 7, 1), date(1940, 8, 1), date(1940, 9, 1), date(1940, 10, 1), date(1940, 11, 1), date(1940, 12, 1), date(1941, 1, 1), date(1941, 2, 1), date(1941, 3, 1), date(1941, 4, 1), date(1941, 5, 1), date(1941, 6, 1), date(1941, 7, 1), date(1941, 8, 1), date(1941, 9, 1), date(1941, 10, 1), date(1941, 11, 1), date(1941, 12, 1), date(1942, 1, 1), date(1942, 2, 1), date(1942, 3, 1), date(1942, 4, 1), date(1942, 5, 1), date(1942, 6, 1), date(1942, 7, 1), date(1942, 8, 1), date(1942, 9, 1), date(1942, 10, 1), date(1942, 11, 1), date(1942, 12, 1), date(1943, 1, 1), date(1943, 2, 1), date(1943, 3, 1), date(1943, 4, 1), date(1943, 5, 1), date(1943, 6, 1), date(1943, 7, 1), date(1943, 8, 1), date(1943, 9, 1), date(1943, 10, 1), date(1943, 11, 1), date(1943, 12, 1), date(1944, 1, 1), date(1944, 2, 1), date(1944, 3, 1), date(1944, 4, 1), date(1944, 5, 1), date(1944, 6, 1), date(1944, 7, 1), date(1944, 8, 1), date(1944, 9, 1), date(1944, 10, 1), date(1944, 11, 1), date(1944, 12, 1), date(1945, 1, 1), date(1945, 2, 1), date(1945, 3, 1), date(1945, 4, 1), date(1945, 5, 1), date(1945, 6, 1), date(1945, 7, 1), date(1945, 8, 1), date(1945, 9, 1), date(1945, 10, 1), date(1945, 11, 1), date(1945, 12, 1), date(1946, 1, 1), date(1946, 2, 1), date(1946, 3, 1), date(1946, 4, 1), date(1946, 5, 1), date(1946, 6, 1), date(1946, 7, 1), date(1946, 8, 1), date(1946, 9, 1), date(1946, 10, 1), date(1946, 11, 1), date(1946, 12, 1), date(1947, 1, 1), date(1947, 2, 1), date(1947, 3, 1), date(1947, 4, 1), date(1947, 5, 1), date(1947, 6, 1), date(1947, 7, 1), date(1947, 8, 1), date(1947, 9, 1), date(1947, 10, 1), date(1947, 11, 1), date(1947, 12, 1), date(1948, 1, 1), date(1948, 2, 1), date(1948, 3, 1), date(1948, 4, 1), date(1948, 5, 1), date(1948, 6, 1), date(1948, 7, 1), date(1948, 8, 1), date(1948, 9, 1), date(1948, 10, 1), date(1948, 11, 1), date(1948, 12, 1), date(1949, 1, 1), date(1949, 2, 1), date(1949, 3, 1), date(1949, 4, 1), date(1949, 5, 1), date(1949, 6, 1), date(1949, 7, 1), date(1949, 8, 1), date(1949, 9, 1), date(1949, 10, 1), date(1949, 11, 1), date(1949, 12, 1), date(1950, 1, 1), date(1950, 2, 1), date(1950, 3, 1), date(1950, 4, 1), date(1950, 5, 1), date(1950, 6, 1), date(1950, 7, 1), date(1950, 8, 1), date(1950, 9, 1), date(1950, 10, 1), date(1950, 11, 1), date(1950, 12, 1), date(1951, 1, 1), date(1951, 2, 1), date(1951, 3, 1), date(1951, 4, 1), date(1951, 5, 1), date(1951, 6, 1), date(1951, 7, 1), date(1951, 8, 1), date(1951, 9, 1), date(1951, 10, 1), date(1951, 11, 1), date(1951, 12, 1), date(1952, 1, 1), date(1952, 2, 1), date(1952, 3, 1), date(1952, 4, 1), date(1952, 5, 1), date(1952, 6, 1), date(1952, 7, 1), date(1952, 8, 1), date(1952, 9, 1), date(1952, 10, 1), date(1952, 11, 1), date(1952, 12, 1), date(1953, 1, 1), date(1953, 2, 1), date(1953, 3, 1), date(1953, 4, 1), date(1953, 5, 1), date(1953, 6, 1), date(1953, 7, 1), date(1953, 8, 1), date(1953, 9, 1), date(1953, 10, 1), date(1953, 11, 1), date(1953, 12, 1), date(1954, 1, 1), date(1954, 2, 1), date(1954, 3, 1), date(1954, 4, 1), date(1954, 5, 1), date(1954, 6, 1), date(1954, 7, 1), date(1954, 8, 1), date(1954, 9, 1), date(1954, 10, 1), date(1954, 11, 1), date(1954, 12, 1), date(1955, 1, 1), date(1955, 2, 1), date(1955, 3, 1), date(1955, 4, 1), date(1955, 5, 1), date(1955, 6, 1), date(1955, 7, 1), date(1955, 8, 1), date(1955, 9, 1), date(1955, 10, 1), date(1955, 11, 1), date(1955, 12, 1), date(1956, 1, 1), date(1956, 2, 1), date(1956, 3, 1), date(1956, 4, 1), date(1956, 5, 1), date(1956, 6, 1), date(1956, 7, 1), date(1956, 8, 1), date(1956, 9, 1), date(1956, 10, 1), date(1956, 11, 1), date(1956, 12, 1), date(1957, 1, 1), date(1957, 2, 1), date(1957, 3, 1), date(1957, 4, 1), date(1957, 5, 1), date(1957, 6, 1), date(1957, 7, 1), date(1957, 8, 1), date(1957, 9, 1), date(1957, 10, 1), date(1957, 11, 1), date(1957, 12, 1), date(1958, 1, 1), date(1958, 2, 1), date(1958, 3, 1), date(1958, 4, 1), date(1958, 5, 1), date(1958, 6, 1), date(1958, 7, 1), date(1958, 8, 1), date(1958, 9, 1), date(1958, 10, 1), date(1958, 11, 1), date(1958, 12, 1), date(1959, 1, 1), date(1959, 2, 1), date(1959, 3, 1), date(1959, 4, 1), date(1959, 5, 1), date(1959, 6, 1), date(1959, 7, 1), date(1959, 8, 1), date(1959, 9, 1), date(1959, 10, 1), date(1959, 11, 1), date(1959, 12, 1), date(1960, 1, 1), date(1960, 2, 1), date(1960, 3, 1), date(1960, 4, 1), date(1960, 5, 1), date(1960, 6, 1), date(1960, 7, 1), date(1960, 8, 1), date(1960, 9, 1), date(1960, 10, 1), date(1960, 11, 1), date(1960, 12, 1), date(1961, 1, 1), date(1961, 2, 1), date(1961, 3, 1), date(1961, 4, 1), date(1961, 5, 1), date(1961, 6, 1), date(1961, 7, 1), date(1961, 8, 1), date(1961, 9, 1), date(1961, 10, 1), date(1961, 11, 1), date(1961, 12, 1), date(1962, 1, 1), date(1962, 2, 1), date(1962, 3, 1), date(1962, 4, 1), date(1962, 5, 1), date(1962, 6, 1), date(1962, 7, 1), date(1962, 8, 1), date(1962, 9, 1), date(1962, 10, 1), date(1962, 11, 1), date(1962, 12, 1), date(1963, 1, 1), date(1963, 2, 1), date(1963, 3, 1), date(1963, 4, 1), date(1963, 5, 1), date(1963, 6, 1), date(1963, 7, 1), date(1963, 8, 1), date(1963, 9, 1), date(1963, 10, 1), date(1963, 11, 1), date(1963, 12, 1), date(1964, 1, 1), date(1964, 2, 1), date(1964, 3, 1), date(1964, 4, 1), date(1964, 5, 1), date(1964, 6, 1), date(1964, 7, 1), date(1964, 8, 1), date(1964, 9, 1), date(1964, 10, 1), date(1964, 11, 1), date(1964, 12, 1), date(1965, 1, 1), date(1965, 2, 1), date(1965, 3, 1), date(1965, 4, 1), date(1965, 5, 1), date(1965, 6, 1), date(1965, 7, 1), date(1965, 8, 1), date(1965, 9, 1), date(1965, 10, 1), date(1965, 11, 1), date(1965, 12, 1), date(1966, 1, 1), date(1966, 2, 1), date(1966, 3, 1), date(1966, 4, 1), date(1966, 5, 1), date(1966, 6, 1), date(1966, 7, 1), date(1966, 8, 1), date(1966, 9, 1), date(1966, 10, 1), date(1966, 11, 1), date(1966, 12, 1), date(1967, 1, 1), date(1967, 2, 1), date(1967, 3, 1), date(1967, 4, 1), date(1967, 5, 1), date(1967, 6, 1), date(1967, 7, 1), date(1967, 8, 1), date(1967, 9, 1), date(1967, 10, 1), date(1967, 11, 1), date(1967, 12, 1), date(1968, 1, 1), date(1968, 2, 1), date(1968, 3, 1), date(1968, 4, 1), date(1968, 5, 1), date(1968, 6, 1), date(1968, 7, 1), date(1968, 8, 1), date(1968, 9, 1), date(1968, 10, 1), date(1968, 11, 1), date(1968, 12, 1), date(1969, 1, 1), date(1969, 2, 1), date(1969, 3, 1), date(1969, 4, 1), date(1969, 5, 1), date(1969, 6, 1), date(1969, 7, 1), date(1969, 8, 1), date(1969, 9, 1), date(1969, 10, 1), date(1969, 11, 1), date(1969, 12, 1), date(1970, 1, 1), date(1970, 2, 1), date(1970, 3, 1), date(1970, 4, 1), date(1970, 5, 1), date(1970, 6, 1), date(1970, 7, 1), date(1970, 8, 1), date(1970, 9, 1), date(1970, 10, 1), date(1970, 11, 1), date(1970, 12, 1), date(1971, 1, 1), date(1971, 2, 1), date(1971, 3, 1), date(1971, 4, 1), date(1971, 5, 1), date(1971, 6, 1), date(1971, 7, 1), date(1971, 8, 1), date(1971, 9, 1), date(1971, 10, 1), date(1971, 11, 1), date(1971, 12, 1), date(1972, 1, 1), date(1972, 2, 1), date(1972, 3, 1), date(1972, 4, 1), date(1972, 5, 1), date(1972, 6, 1), date(1972, 7, 1), date(1972, 8, 1), date(1972, 9, 1), date(1972, 10, 1), date(1972, 11, 1), date(1972, 12, 1), date(1973, 1, 1), date(1973, 2, 1), date(1973, 3, 1), date(1973, 4, 1), date(1973, 5, 1), date(1973, 6, 1), date(1973, 7, 1), date(1973, 8, 1), date(1973, 9, 1), date(1973, 10, 1), date(1973, 11, 1), date(1973, 12, 1), date(1974, 1, 1), date(1974, 2, 1), date(1974, 3, 1), date(1974, 4, 1), date(1974, 5, 1), date(1974, 6, 1), date(1974, 7, 1), date(1974, 8, 1), date(1974, 9, 1), date(1974, 10, 1), date(1974, 11, 1), date(1974, 12, 1), date(1975, 1, 1), date(1975, 2, 1), date(1975, 3, 1), date(1975, 4, 1), date(1975, 5, 1), date(1975, 6, 1), date(1975, 7, 1), date(1975, 8, 1), date(1975, 9, 1), date(1975, 10, 1), date(1975, 11, 1), date(1975, 12, 1), date(1976, 1, 1), date(1976, 2, 1), date(1976, 3, 1), date(1976, 4, 1), date(1976, 5, 1), date(1976, 6, 1), date(1976, 7, 1), date(1976, 8, 1), date(1976, 9, 1), date(1976, 10, 1), date(1976, 11, 1), date(1976, 12, 1), date(1977, 1, 1), date(1977, 2, 1), date(1977, 3, 1), date(1977, 4, 1), date(1977, 5, 1), date(1977, 6, 1), date(1977, 7, 1), date(1977, 8, 1), date(1977, 9, 1), date(1977, 10, 1), date(1977, 11, 1), date(1977, 12, 1), date(1978, 1, 1), date(1978, 2, 1), date(1978, 3, 1), date(1978, 4, 1), date(1978, 5, 1), date(1978, 6, 1), date(1978, 7, 1), date(1978, 8, 1), date(1978, 9, 1), date(1978, 10, 1), date(1978, 11, 1), date(1978, 12, 1), date(1979, 1, 1), date(1979, 2, 1), date(1979, 3, 1), date(1979, 4, 1), date(1979, 5, 1), date(1979, 6, 1), date(1979, 7, 1), date(1979, 8, 1), date(1979, 9, 1), date(1979, 10, 1), date(1979, 11, 1), date(1979, 12, 1), date(1980, 1, 1), date(1980, 2, 1), date(1980, 3, 1), date(1980, 4, 1), date(1980, 5, 1), date(1980, 6, 1), date(1980, 7, 1), date(1980, 8, 1), date(1980, 9, 1), date(1980, 10, 1), date(1980, 11, 1), date(1980, 12, 1), date(1981, 1, 1), date(1981, 2, 1), date(1981, 3, 1), date(1981, 4, 1), date(1981, 5, 1), date(1981, 6, 1), date(1981, 7, 1), date(1981, 8, 1), date(1981, 9, 1), date(1981, 10, 1), date(1981, 11, 1), date(1981, 12, 1), date(1982, 1, 1), date(1982, 2, 1), date(1982, 3, 1), date(1982, 4, 1), date(1982, 5, 1), date(1982, 6, 1), date(1982, 7, 1), date(1982, 8, 1), date(1982, 9, 1), date(1982, 10, 1), date(1982, 11, 1), date(1982, 12, 1), date(1983, 1, 1), date(1983, 2, 1), date(1983, 3, 1), date(1983, 4, 1), date(1983, 5, 1), date(1983, 6, 1), date(1983, 7, 1), date(1983, 8, 1), date(1983, 9, 1), date(1983, 10, 1), date(1983, 11, 1), date(1983, 12, 1), date(1984, 1, 1), date(1984, 2, 1), date(1984, 3, 1), date(1984, 4, 1), date(1984, 5, 1), date(1984, 6, 1), date(1984, 7, 1), date(1984, 8, 1), date(1984, 9, 1), date(1984, 10, 1), date(1984, 11, 1), date(1984, 12, 1), date(1985, 1, 1), date(1985, 2, 1), date(1985, 3, 1), date(1985, 4, 1), date(1985, 5, 1), date(1985, 6, 1), date(1985, 7, 1), date(1985, 8, 1), date(1985, 9, 1), date(1985, 10, 1), date(1985, 11, 1), date(1985, 12, 1), date(1986, 1, 1), date(1986, 2, 1), date(1986, 3, 1), date(1986, 4, 1), date(1986, 5, 1), date(1986, 6, 1), date(1986, 7, 1), date(1986, 8, 1), date(1986, 9, 1), date(1986, 10, 1), date(1986, 11, 1), date(1986, 12, 1), date(1987, 1, 1), date(1987, 2, 1), date(1987, 3, 1), date(1987, 4, 1), date(1987, 5, 1), date(1987, 6, 1), date(1987, 7, 1), date(1987, 8, 1), date(1987, 9, 1), date(1987, 10, 1), date(1987, 11, 1), date(1987, 12, 1), date(1988, 1, 1), date(1988, 2, 1), date(1988, 3, 1), date(1988, 4, 1), date(1988, 5, 1), date(1988, 6, 1), date(1988, 7, 1), date(1988, 8, 1), date(1988, 9, 1), date(1988, 10, 1), date(1988, 11, 1), date(1988, 12, 1), date(1989, 1, 1), date(1989, 2, 1), date(1989, 3, 1), date(1989, 4, 1), date(1989, 5, 1), date(1989, 6, 1), date(1989, 7, 1), date(1989, 8, 1), date(1989, 9, 1), date(1989, 10, 1), date(1989, 11, 1), date(1989, 12, 1), date(1990, 1, 1), date(1990, 2, 1), date(1990, 3, 1), date(1990, 4, 1), date(1990, 5, 1), date(1990, 6, 1), date(1990, 7, 1), date(1990, 8, 1), date(1990, 9, 1), date(1990, 10, 1), date(1990, 11, 1), date(1990, 12, 1), date(1991, 1, 1), date(1991, 2, 1), date(1991, 3, 1), date(1991, 4, 1), date(1991, 5, 1), date(1991, 6, 1), date(1991, 7, 1), date(1991, 8, 1), date(1991, 9, 1), date(1991, 10, 1), date(1991, 11, 1), date(1991, 12, 1), date(1992, 1, 1), date(1992, 2, 1), date(1992, 3, 1), date(1992, 4, 1), date(1992, 5, 1), date(1992, 6, 1), date(1992, 7, 1), date(1992, 8, 1), date(1992, 9, 1), date(1992, 10, 1), date(1992, 11, 1), date(1992, 12, 1), date(1993, 1, 1), date(1993, 2, 1), date(1993, 3, 1), date(1993, 4, 1), date(1993, 5, 1), date(1993, 6, 1), date(1993, 7, 1), date(1993, 8, 1), date(1993, 9, 1), date(1993, 10, 1), date(1993, 11, 1), date(1993, 12, 1), date(1994, 1, 1), date(1994, 2, 1), date(1994, 3, 1), date(1994, 4, 1), date(1994, 5, 1), date(1994, 6, 1), date(1994, 7, 1), date(1994, 8, 1), date(1994, 9, 1), date(1994, 10, 1), date(1994, 11, 1), date(1994, 12, 1), date(1995, 1, 1), date(1995, 2, 1), date(1995, 3, 1), date(1995, 4, 1), date(1995, 5, 1), date(1995, 6, 1), date(1995, 7, 1), date(1995, 8, 1), date(1995, 9, 1), date(1995, 10, 1), date(1995, 11, 1), date(1995, 12, 1), date(1996, 1, 1), date(1996, 2, 1), date(1996, 3, 1), date(1996, 4, 1), date(1996, 5, 1), date(1996, 6, 1), date(1996, 7, 1), date(1996, 8, 1), date(1996, 9, 1), date(1996, 10, 1), date(1996, 11, 1), date(1996, 12, 1), date(1997, 1, 1), date(1997, 2, 1), date(1997, 3, 1), date(1997, 4, 1), date(1997, 5, 1), date(1997, 6, 1), date(1997, 7, 1), date(1997, 8, 1), date(1997, 9, 1), date(1997, 10, 1), date(1997, 11, 1), date(1997, 12, 1), date(1998, 1, 1), date(1998, 2, 1), date(1998, 3, 1), date(1998, 4, 1), date(1998, 5, 1), date(1998, 6, 1), date(1998, 7, 1), date(1998, 8, 1), date(1998, 9, 1), date(1998, 10, 1), date(1998, 11, 1), date(1998, 12, 1), date(1999, 1, 1), date(1999, 2, 1), date(1999, 3, 1), date(1999, 4, 1), date(1999, 5, 1), date(1999, 6, 1), date(1999, 7, 1), date(1999, 8, 1), date(1999, 9, 1), date(1999, 10, 1), date(1999, 11, 1), date(1999, 12, 1), date(2000, 1, 1), date(2000, 2, 1), date(2000, 3, 1), date(2000, 4, 1), date(2000, 5, 1), date(2000, 6, 1), date(2000, 7, 1), date(2000, 8, 1), date(2000, 9, 1), date(2000, 10, 1), date(2000, 11, 1), date(2000, 12, 1), date(2001, 1, 1), date(2001, 2, 1), date(2001, 3, 1), date(2001, 4, 1), date(2001, 5, 1), date(2001, 6, 1), date(2001, 7, 1), date(2001, 8, 1), date(2001, 9, 1), date(2001, 10, 1), date(2001, 11, 1), date(2001, 12, 1), date(2002, 1, 1), date(2002, 2, 1), date(2002, 3, 1), date(2002, 4, 1), date(2002, 5, 1), date(2002, 6, 1), date(2002, 7, 1), date(2002, 8, 1), date(2002, 9, 1), date(2002, 10, 1), date(2002, 11, 1), date(2002, 12, 1), date(2003, 1, 1), date(2003, 2, 1), date(2003, 3, 1), date(2003, 4, 1), date(2003, 5, 1), date(2003, 6, 1), date(2003, 7, 1), date(2003, 8, 1), date(2003, 9, 1), date(2003, 10, 1), date(2003, 11, 1), date(2003, 12, 1), date(2004, 1, 1), date(2004, 2, 1), date(2004, 3, 1), date(2004, 4, 1), date(2004, 5, 1), date(2004, 6, 1), date(2004, 7, 1), date(2004, 8, 1), date(2004, 9, 1), date(2004, 10, 1), date(2004, 11, 1), date(2004, 12, 1), date(2005, 1, 1), date(2005, 2, 1), date(2005, 3, 1), date(2005, 4, 1), date(2005, 5, 1), date(2005, 6, 1), date(2005, 7, 1), date(2005, 8, 1), date(2005, 9, 1), date(2005, 10, 1), date(2005, 11, 1), date(2005, 12, 1), date(2006, 1, 1), date(2006, 2, 1), date(2006, 3, 1), date(2006, 4, 1), date(2006, 5, 1), date(2006, 6, 1), date(2006, 7, 1), date(2006, 8, 1), date(2006, 9, 1), date(2006, 10, 1), date(2006, 11, 1), date(2006, 12, 1), date(2007, 1, 1), date(2007, 2, 1), date(2007, 3, 1), date(2007, 4, 1), date(2007, 5, 1), date(2007, 6, 1), date(2007, 7, 1), date(2007, 8, 1), date(2007, 9, 1), date(2007, 10, 1), date(2007, 11, 1), date(2007, 12, 1), date(2008, 1, 1), date(2008, 2, 1), date(2008, 3, 1), date(2008, 4, 1), date(2008, 5, 1), date(2008, 6, 1), date(2008, 7, 1), date(2008, 8, 1), date(2008, 9, 1), date(2008, 10, 1), date(2008, 11, 1), date(2008, 12, 1), date(2009, 1, 1), date(2009, 2, 1), date(2009, 3, 1), date(2009, 4, 1), date(2009, 5, 1), date(2009, 6, 1), date(2009, 7, 1), date(2009, 8, 1), date(2009, 9, 1), date(2009, 10, 1), date(2009, 11, 1), date(2009, 12, 1), date(2010, 1, 1), date(2010, 2, 1), date(2010, 3, 1), date(2010, 4, 1), date(2010, 5, 1), date(2010, 6, 1), date(2010, 7, 1), date(2010, 8, 1), date(2010, 9, 1), date(2010, 10, 1), date(2010, 11, 1), date(2010, 12, 1), date(2011, 1, 1), date(2011, 2, 1), date(2011, 3, 1), date(2011, 4, 1), date(2011, 5, 1), date(2011, 6, 1), date(2011, 7, 1), date(2011, 8, 1), date(2011, 9, 1), date(2011, 10, 1), date(2011, 11, 1), date(2011, 12, 1), date(2012, 1, 1), date(2012, 2, 1), date(2012, 3, 1), date(2012, 4, 1), date(2012, 5, 1), date(2012, 6, 1), date(2012, 7, 1), date(2012, 8, 1), date(2012, 9, 1), date(2012, 10, 1), date(2012, 11, 1), date(2012, 12, 1), date(2013, 1, 1), date(2013, 2, 1), date(2013, 3, 1), date(2013, 4, 1), date(2013, 5, 1), date(2013, 6, 1), date(2013, 7, 1), date(2013, 8, 1), date(2013, 9, 1), date(2013, 10, 1), date(2013, 11, 1), date(2013, 12, 1), date(2014, 1, 1), date(2014, 2, 1), date(2014, 3, 1), date(2014, 4, 1), date(2014, 5, 1), date(2014, 6, 1), date(2014, 7, 1), date(2014, 8, 1), date(2014, 9, 1), date(2014, 10, 1), date(2014, 11, 1), date(2014, 12, 1), date(2015, 1, 1), date(2015, 2, 1), date(2015, 3, 1), date(2015, 4, 1), date(2015, 5, 1), date(2015, 6, 1), date(2015, 7, 1), date(2015, 8, 1), date(2015, 9, 1), date(2015, 10, 1), date(2015, 11, 1), date(2015, 12, 1), date(2016, 1, 1), date(2016, 2, 1), date(2016, 3, 1), date(2016, 4, 1), date(2016, 5, 1), date(2016, 6, 1), date(2016, 7, 1), date(2016, 8, 1), date(2016, 9, 1), date(2016, 10, 1), date(2016, 11, 1), date(2016, 12, 1), date(2017, 1, 1), date(2017, 2, 1), date(2017, 3, 1), date(2017, 4, 1), date(2017, 5, 1), date(2017, 6, 1), date(2017, 7, 1), date(2017, 8, 1), date(2017, 9, 1), date(2017, 10, 1), date(2017, 11, 1), date(2017, 12, 1), date(2018, 1, 1), date(2018, 2, 1), date(2018, 3, 1)]
        self.data = [[x, 1] for x in d]
        return super().setUp()
    
    def test_single(self):
        want = [[date(1871, 1, 1), 1]]
        self.assertEqual(get_investment_window(date(1871, 1, 1), date(1871, 1, 1), self.data), want)

    def test_range(self):
        want = [[date(1871, 1, 1), 1], [date(1871, 2, 1), 1], [date(1871, 3, 1), 1]]
        self.assertEqual(get_investment_window(date(1871, 1, 1), date(1871, 3, 1), self.data), want)

    def test_out_of_range(self):
        want = [[date(1870, 1, 1), 1], [date(1871, 2, 1), 1], [date(1871, 3, 1), 1]]
        self.assertIsNone(get_investment_window(date(1870, 1, 1), date(1871, 3, 1), self.data))
        self.assertIsNone(get_investment_window(date(1871, 1, 1), date(2080, 3, 1), self.data))
        self.assertIsNone(get_investment_window(date(2080, 3, 1), date(2090, 3, 1), self.data))

    def test_broken(self):
        start = date(1880, 2, 1)
        end = date(1881, 2, 1)
        want = [[get_duration_end_date(start, months=nmo), 1] for nmo in range(13)]
        self.assertEqual(get_investment_window(start, end, self.data), want)
    
    def test_closeby(self):
        start, end = date(1871, 12, 1), date(1882, 1, 1)
        want = [[get_duration_end_date(start, months=nmo), 1] for nmo in range(122)]
        self.assertEqual(get_investment_window(start, end, self.data), want)
    
    def test_count_months_delta(self):
        start, end = date(1871, 12, 1), date(1882, 1, 1)
        self.assertEqual(count_months_delta(start, end), 121)


class TestClusterData(unittest.TestCase):
    def test_foo(self):
        self.assertEqual(smart_round(10), 10)
        self.assertEqual(smart_round(1), 1)
        self.assertEqual(smart_round(1.1111), 1.11)
        self.assertEqual(smart_round(1000212123.1111), 1000000000)
        self.assertEqual(smart_round(2000212123.1111), 2000000000)

    def test_large(self):
        self.assertEqual(smart_round(1.1111, significant_digits=3), 1.11)
        self.assertEqual(smart_round(123, significant_digits=3), 123)
        self.assertEqual(smart_round(1234.00, significant_digits=3), 1230)